/*
  Dynamixel Servo driver for use on a Digilent ChipKIT board
  using MPIDE Arduino IDE platform.
  
  Version 2 - cleaned up syntax quite a bit, fixed bugs
  
  (c) Dennis Clark May 2013
  You are free to use this any way you want.
 */
 
//#define _DEBUG_

/*
 *  Assignments
*/
#define DLED 16
#define nTxU2 19

// Dynamixel commands
#define DM_PING       0x01
#define DM_READ_DATA  0x02
#define DM_WRITE_DATA 0x03
#define DM_REG_WRITE  0x04
#define DM_ACTION     0x05
#define DM_RESET      0x06
#define DM_SYNC_WRITE 0x83

// Control Table memory addresses and parameter bytes
uint8_t  ct[25] = {3,4,5,6,8,11,12,13,14,16,17,18,19,24,25,26,27,28,29,30,32,34,44,47,48};
uint8_t  cb[25] = {1,1,1,2,2,1 ,1, 1 ,2 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,2 ,2 ,2 ,1 ,1 ,2};

// Command Table IDs
typedef enum DMControls_e
{
  dmID =0,
  dmBaud,
  dmRetDelay,
  dmCWAngleLimit,
  dmCCWAngleLimit,
  dmHighTempLimit,
  dmLowVoltLimit,
  dmHighVoltLimit,
  dmMaxTorque,
  dmStatRetLevel,
  dmAlarmLED,
  dmAlarmShutdn,
  dmReserved,
  dmTorqueEn,
  dmLED,
  dmCWCompMargin,
  dmCCWCompMargin,
  dmCWCompSlope,
  dmCCWCompSlope,
  dmPosition,
  dmMovSpeed,
  dmTorqueLimit,
  dmRegInst,
  dmLock,
  dmPunch
} DMcontrols_e;

uint8_t err;
uint8_t stat[10], statLen;

#define LEFT 19
#define RIGHT 20
#define MOTOR 0
#define SERVO 1023
#define FWD 0
#define BWD 1

void setup() {
  uint8_t n;
  
  pinMode(nTxU2, OUTPUT);    // NOT Tx/Rx: low means transmit, high means receive
  pinMode(DLED, OUTPUT);     // Das Blinken Lights, just to have an LED
  
  Serial1.begin(1000000);    // Talk to Dynamixel AX12a at 1Mbps
  Serial.begin(115200);      // Talk to computer at 115.2Kbps
    
  //LookBus(20);
  
  /*
  * Set servo speed to full (1023)
  */
  err = CmdSpeed(LEFT, 0);
  delay(5);
  err = CmdSpeed(RIGHT, 0);

  /*
  * Read back our servo speed
  */
  err = CmdReadControl(LEFT, dmMovSpeed, stat, &statLen);
  Serial.print("LEFT servo speed "); Serial.println((stat[5] + stat[6]*256));  
  err = CmdReadControl(RIGHT, dmMovSpeed, stat, &statLen);
  Serial.print("RIGHT servo speed "); Serial.println((stat[5] + stat[6]*256));
  Serial.println("");
  
  Serial.println("Ready to roll\n");
}
  
void loop() 
{
  uint8_t n;
  uint16_t m;  
  
  err = CmdLED(LEFT, 1);
  delay(5);
  err = CmdLED(RIGHT, 1);
  delay(5);
  Serial.println("Set turnmode");
  err = CmdTurnMode(LEFT,RIGHT,MOTOR);
  /*
  *  We appear to need a delay after this or our
  *  first motor command will not be responded to.
  */
  delay(100);
  Serial.println("Stop Motors");
  err = CmdRoll(LEFT, FWD, 0, RIGHT, FWD, 0);
  PrintError(err);
  delay(1000);
  Serial.println("fwd 1");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, FWD, 511);  
  PrintError(err);
  delay(2000);
  Serial.println("Turn 1");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, BWD, 511); 
  PrintError(err);
  delay(2000);
  Serial.println("FWD 2");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, FWD, 511);
  PrintError(err);
  delay(2000);
  Serial.println("Turn 2");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, BWD, 511); 
  PrintError(err);
  delay(2000);
  Serial.println("FWD 3");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, FWD, 511); 
  PrintError(err);
  delay(2000);
  Serial.println("Turn 3");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, BWD, 511);  
  PrintError(err);
  delay(3000);
  Serial.println("FWD 4");
  err = CmdRoll(LEFT, FWD, 511, RIGHT, FWD, 511);  
  PrintError(err);
  delay(2000);
  Serial.println("Stop all");
  err = CmdRoll(LEFT, FWD, 0, RIGHT, FWD, 0);
  PrintError(err);
  
  err = CmdLED(LEFT, 0);
  delay(5);
  err = CmdLED(RIGHT, 0);
  
  while(1);
}

uint8_t CmdTurnMode(uint8_t lSid, uint8_t rSid, uint16_t mode)
/*
*  Make these two servos continuous rotation.
*  mode: 0 = cont. rotation, 1023 = servo
*  Returns error status
*/
{
  uint8_t data[2];
  uint8_t err;
  
  data[0] = (uint8_t)(mode);
  data[1] = (uint8_t)(mode >>8);
  
  err  = CmdWriteControl(lSid,data,dmCCWAngleLimit);
  err |= CmdWriteControl(rSid,data,dmCCWAngleLimit);
  
  return(err);
}  

uint8_t CmdRoll(uint8_t lSid, uint8_t lDir, uint16_t lSpd, 
                uint8_t rSid, uint8_t rDir, uint16_t rSpd)
/*
* Make robot go in a certain direction: 
* [l/r]Sid = left/right servo ID
* [l/r]Dir = 0 CCW, 1 CW
* [l/r]Spd = speed 0-1023
* Returns error status
*/
{
  uint8_t data[2];
  uint8_t err;
  
  /*
  * Servos reversion direction when mirrored, so reverse 
  * the direction bit on the right side servo.
  */
  rDir = ~rDir&0x01;
  
  /*
  * direction and speed are in a 16 bit data word:
  *  15-11: 00000 | 10: [0=CCW, 1=CW] | 9-0: speed value (0-1023)
  */
  data[0] = (uint8_t)lSpd;
  data[1] = (uint8_t)(4*lDir) | (uint8_t)((lSpd >>8)&0xFC);
  err  = CmdWriteControl(lSid,data,dmMovSpeed);
  
  data[0] = (uint8_t)rSpd;
  data[1] = (uint8_t)(4*rDir) | (uint8_t)((rSpd >>8)&0xFC);
  err |= CmdWriteControl(rSid,data,dmMovSpeed);  
  
  return(err);
}

uint8_t CmdReset(uint8_t sid)
/*
*  Send a reset (may be global)
*  Clears all paramters in the servo to default values.
*/
{
  uint8_t cmd[6], cmdLen;
  uint8_t err;
  
  cmd[0] = 0xFF;
  cmd[1] = 0xFF;
  cmd[2] = sid;
  cmd[3] = 0x02;
  cmd[4] = DM_RESET;
  cmd[5] = GetCSUM(cmd,3);  // get checksum
  
  cmdLen = 6;
  err = SendCmd(cmd, cmdLen, stat, &statLen);
  return(err);
}  

uint8_t CmdPing(uint8_t sid, uint8_t *stat, uint8_t *statLen)
/*
*  Look for a servo at a specific address.
*  return error status
*/
{
  uint8_t cmd[6], cmdLen;
  uint8_t err;
  
  cmd[0] = 0xFF;
  cmd[1] = 0xFF;
  cmd[2] = sid;
  cmd[3] = 0x02;
  cmd[4] = DM_PING;
  cmd[5] = GetCSUM(cmd,3);  // get checksum
  
  cmdLen = 6;
  err = SendCmd(cmd, cmdLen, stat, statLen);
  return(err);
}

void LookBus(uint8_t limit)
/*
* Scan the bus from address 0 to 'limit' to find all servos there.
*/
{
  uint8_t n;
  uint8_t stat[6],statLen;
  
  for (n=0; n<=limit; n++)
  {
    err = CmdPing(n, stat, &statLen);
    if (err != 0xFF)
    {
      Serial.print("Stat: ");Serial.print(n,DEC);Serial.print("= ");PrintStatus(stat, statLen);
    }
    delay(100);
  }
}

uint8_t CmdLED(uint8_t sid, uint8_t val)
/*
*  Turn an LED on or off on the AX12a
*  sid = the servo ID
*  val = 0(off), 1(on)
*
*  error status
*/
{ 
  uint8_t data[1];
  uint8_t err;
  
  data[0] = val;
  err = CmdWriteControl(sid,data,dmLED);
  return(err);
}

uint8_t CmdMove(uint8_t sid, uint16_t val)
/*
*  move the servo motor to a position.
*  sid = the servo ID
*  val = the position
*
*  Returns error status
*/
{
  uint8_t data[2];
  uint8_t err;
  
  data[0] = (uint8_t)val;
  data[1] = (uint8_t)(val >>8);
  err = CmdWriteControl(sid, data, dmPosition);
  return(err);
}

uint8_t CmdSpeed(uint8_t sid, uint16_t val)
/*
*  move the servo motor to a position.
*  sid = the servo ID
*  val = 0(off), 1(on)
*
*  Returns error status
*/
{
  uint8_t data[2];
  uint8_t err;
  
  data[0] = (uint8_t)val;
  data[1] = (uint8_t)(val >>8);
  err = CmdWriteControl(sid, data, dmMovSpeed);
  return(err);
}

uint8_t CmdWriteControl(uint8_t sid, uint8_t *data, uint8_t dm)
/*
*  Format a write to a Dynamixel Control Table.
*  sid = servo ID
*  data = the parameter data (1 or 2 bytes)
*  dm = Control Address
*  returns status of the command sent
*/
{
  uint8_t n;
  uint8_t cmd[10];
  uint8_t cmdLen;
  uint8_t stat[10];
  uint8_t statLen;
  uint8_t err;
  
  cmd[0] = 0xFF;           // sync byte
  cmd[1] = 0xFF;           // sync byte
  cmd[2] = sid;            // ID
  cmd[3] = 3+cb[dm];       // sid+instruction+#params+param bytes
  cmd[4] = DM_WRITE_DATA;  // instruction
  cmd[5] = ct[dm];         // # of params
  for (n=0;n<cb[dm];n++) { // parameters
    cmd[6+n] = data[n];
  }
  cmd[6+cb[dm]] = GetCSUM(cmd, 4+cb[dm]);
  cmdLen = 7+cb[dm];
  
  err = SendCmd(cmd, cmdLen, stat, &statLen);
  return(err);
}

uint8_t CmdReadControl(uint8_t sid, uint8_t addr, uint8_t *stat, uint8_t *statLen)
/*
*  Read out a piece of the Dynamixel Control Table.
*  sid = servo ID
*  addr = the address in the Control Table to read
*  stat = status (values)
*  statLen = the length of data read back
*  return error status
*/
{
  uint8_t cmd[10];
  uint8_t cmdLen;
  
  cmd[0] = 0xFF;
  cmd[1] = 0xFF;
  cmd[2] = sid;
  cmd[3] = 4;
  cmd[4] = DM_READ_DATA;
  cmd[5] = ct[addr];
  cmd[6] = cb[addr];
  cmd[7] = GetCSUM(cmd,5);  
  
  cmdLen = 8;
  err = SendCmd(cmd, cmdLen, stat, statLen);
  return(err);
}

uint8_t GetCSUM(uint8_t *cmd, uint8_t len)
/*
*  calculate the Robotis csum by adding all the
*  command bytes together and taking their binary
*  inverse.  Bytes 0 and 1 are always 0xFF, don't
*  count them.  We need every byte except the first
*  two 0xFF bytes and the CSUM byte itself counted.
*/
{
  uint16_t n;
  uint8_t csum = 0;
  
  for (n=2; n<(len+2); n++)
  {
    csum+= cmd[n];
  }
  csum = ~csum;
  
  return(csum);
}

uint8_t SendCmd(uint8_t *cmd, uint8_t cLen, uint8_t *stat, uint8_t *sLen)
/*
*  Send Dynamixel command, return status in the cmd array
*  INPUT: cmd[] to send , cLen is length of command
*  OUPUT: stat[] status returned , sLen is length of status
*
*  RETURNS: non-zero if an error
*/
{
  uint8_t n = 0;
  uint32_t t;

#ifdef _DEBUG_
  Serial.print(cmd[2],DEC);Serial.print(":CMD= ");PrintStatus(cmd, cLen);
#endif

  digitalWrite(nTxU2, LOW);    // writing out cmd
  Serial1.write(cmd, cLen);

  while(U2STAbits.TRMT == 0);    // Wait for all bytes to go out
  digitalWrite(U2MODEbits.ON, 0);// Turn UART off
  digitalWrite(nTxU2, HIGH);     // switch to listen
  Serial1.read();                // the switch generates a false start bit
  digitalWrite(U2MODEbits.ON,1); // Turn UART back on
  
  t = micros() + 500;            // This timer not very accurate
  while (micros() < t)           // wait for reception to start
  {
    if (Serial1.available())
    {
      stat[0] = Serial1.read();
      if (stat[0] = 0xFF)
      {
        n++;
        break;
      }
    }
  }
  
  if (n == 0)                  // never got start byte
  {
    return(0xFF);              // error flag
    *sLen = 0;
  }
  
  t = micros() + 500;
  while(micros() < t)          // Get the rest of the packet
  {
      while(Serial1.available() > 0)
      {
        stat[n++] = Serial1.read();
      }
  }
  *sLen = n;                    // what we read back, always seem to get one extra...
  
#ifdef _DEBUG_
  Serial.print(cmd[2],DEC);Serial.print(":Stat= ");PrintStatus(stat, *sLen);
#endif

  return(stat[4]);              // return the error byte
}

void PrintStatus(uint8_t *stat, uint8_t sLen)
{
  uint8_t n;
  
  for (n=0; n<sLen; n++)
  {
    Serial.print(stat[n], HEX);Serial.print(" ");
  }
  Serial.println("");
}

void PrintError(uint8_t err)
/*
*  Print out the meaning of any error received from a Dynamixel servo.
*/
{
  if (err == 0) 
  {
    return;
  }
  if (err == 0xFF) 
  {
    Serial.println("Servo did not answer");
    return;
  }
  if (err & 0x01) Serial.println("Input Voltage Error");
  if (err & 0x02) Serial.println("Angle Limit Error");
  if (err & 0x04) Serial.println("Overheating Error");
  if (err & 0x08) Serial.println("Range Error");
  if (err & 0x10) Serial.println("Checksum Error");
  if (err & 0x20) Serial.println("Overload Error");
  if (err & 0x40) Serial.println("Instruction Error");
  Serial.print("Status ");Serial.println(err,HEX);
}


